import pandas as pd
from tqdm import tqdm
from CDmetrics.utils import tune_parameters
import numpy as np

from sklearn.model_selection import KFold
from CDmetrics.nn import NN


def compute_metric(data, num_folds, target_column, max_layers, max_units, resources):

    kfold = KFold(n_splits=num_folds, shuffle=True, random_state=0)
    fold_order = [
        [(i + j) % num_folds for i in range(num_folds)] for j in range(num_folds)
    ]
    fold_index = [test_index for _, test_index in kfold.split(data)]
    difficulity = []
    difficulity_index = []

    for folds in tqdm(fold_order):

        train_data_index = []
        for index in folds[0 : len(fold_order) // 2]:
            train_data_index += list(fold_index[index])

        evaluation_data_index = []
        for index in folds[len(fold_order) // 2 : len(fold_order) - 1]:
            evaluation_data_index += list(fold_index[index])
        test_data_index = fold_index[folds[len(fold_order) - 1]]
        train_data = data.iloc[train_data_index]
        evaluation_data = data.iloc[evaluation_data_index]
        test_data = data.iloc[test_data_index]
        best_model_A = NN(
            tune_parameters(
                NN.tune, train_data, target_column, max_layers, max_units, resources
            )
        ).model
        evaluation_data_x = evaluation_data.drop(columns=[target_column], axis=1)
        evaluation_data_y = evaluation_data[target_column]
        best_model_A_predictions = best_model_A.predict(evaluation_data_x)
        evaluation_data[target_column] = pd.DataFrame(
            (
                np.array(best_model_A_predictions.reshape(-1))
                != np.array(evaluation_data_y)
            )
            * 1,
            index=evaluation_data_x.index,
        )
        best_model_B = NN(
            tune_parameters(
                NN.tune,
                evaluation_data,
                target_column,
                max_layers,
                max_units,
                resources,
            )
        ).model
        test_data_x = test_data.drop(columns=[target_column], axis=1)
        predicted_difficulty = 1 - best_model_B.predict(test_data_x, verbose=0).reshape(
            -1
        )
        difficulity.extend(predicted_difficulty)
        difficulity_index.extend(test_data_index)
    return pd.DataFrame(difficulity, index=difficulity_index)



from ray.air.integrations.keras import ReportCheckpointCallback

import tensorflow as tf
from keras.callbacks import EarlyStopping


class NN:
    def __init__(self, params) -> None:
        self.num_classes = params["num_classes"]

        self.batch_size = params["batch_size"]
        self.params = params

        # First hidden layer with input shape
        self.model = tf.keras.models.Sequential()
        self.model.add(
            tf.keras.layers.Dense(
                params["number_of_neurons"][0],
                input_shape=(params["input_size"],),
                activation=params["activation"],
            )
        )
        # Second hidden layer to number of hidden layers
        for i in range(1, len(params["number_of_neurons"])):
            self.model.add(
                tf.keras.layers.Dense(
                    params["number_of_neurons"][i], activation=params["activation"]
                )
            )

        if self.num_classes > 2:
            self.loss_function = "categorical_crossentropy"
            self.output_activation = "softmax"
        else:
            self.num_classes = 1
            self.loss_function = "binary_crossentropy"
            self.output_activation = "sigmoid"
            self.output_activation = "softmax"

        # Ouput layer
        self.model.add(
            tf.keras.layers.Dense(self.num_classes, activation=self.output_activation)
        )
        self.model.compile(
            loss=self.loss_function,
            optimizer=tf.keras.optimizers.Adam(learning_rate=params["learnRate"]),
            metrics=["accuracy"],
        )

    def tune(config, data_x, data_y):
        """ """
        network = NN(config)
        model = network.model
        es = EarlyStopping(monitor="val_loss", mode="min", verbose=1, patience=30)
        model.fit(
            data_x,
            data_y,
            verbose=0,
            validation_split=0.3,
            batch_size=config["batch_size"],
            epochs=100,
            callbacks=[
                es,
                ReportCheckpointCallback(metrics={"mean_accuracy": "accuracy"}),
            ],
        )

    def train(self, data_x, data_y):
        """ """
        es = EarlyStopping(monitor="val_loss", mode="min", verbose=1, patience=30)
        self.model.fit(
            data_x,
            data_y,
            verbose=0,
            validation_split=0.3,
            batch_size=32,
            epochs=100,
            callbacks=[
                es,
                ReportCheckpointCallback(metrics={"mean_accuracy": "accuracy"}),
            ],
        )

        return self.model

    def predict(X_test):
        """ """



import numpy as np
import pandas as pd
from CDmetrics.nn import NN


def compute_metric(data, number_of_NNs, target_column):
    max_hidden_units = round(len(data) * 0.01)
    Threshold = number_of_NNs * 0.9

    X_overall = data.drop(columns=[target_column], axis=1)
    y_overall = data[target_column]

    difficulity = []
    for index in range(len(data)):
        X_test = X_overall.iloc[[index]]
        y_test = y_overall[index]

        X = X_overall.drop(index=[index])
        y = y_overall.drop(index=[index])

        params = {}
        params.update(
            {
                "learnRate": 0.01,
                "batch_size": 32,
                "activation": "relu",
                "num_classes": len(set(y.values)),
                "input_size": len(X.columns),
            }
        )

        n_neureons_hidden_layer = 0
        count = 0
        while n_neureons_hidden_layer < max_hidden_units:
            if count < Threshold:
                n_neureons_hidden_layer += 1
                count = 0
                for _ in range(number_of_NNs):
                    params.update({"number_of_neurons": [n_neureons_hidden_layer]})
                    model = NN(params)
                    trained_model = model.train(X, y)
                    y_test = np.argmax(y_test)
                    y_pred = np.argmax(trained_model.predict(np.array(X_test)), axis=1)
                    if np.all(y_pred == y_test):
                        count += 1
                    else:
                        count += 0
            else:
                difficulity.append(n_neureons_hidden_layer / max_hidden_units)
                break

    return pd.DataFrame(difficulity)



import pandas as pd
from tqdm import tqdm
import numpy as np
import math
from CDmetrics.nn import NN
from CDmetrics.utils import tune_parameters


def compute_metric(
    data, target_column, number_of_predictions, max_layers, max_units, resources
):
    difficulity = []
    for index in tqdm(range(len(data))):
        test_data = data.iloc[[index]]
        test_data_x = test_data.drop(columns=[target_column], axis=1)
        test_data_y = test_data[target_column]
        train_data = data.drop(index=index, axis=0)

        best_config = tune_parameters(
            NN.tune, train_data, target_column, max_layers, max_units, resources
        )
        best_model = NN(best_config)
        predictions = []
        for i in range(number_of_predictions):
            prediction = best_model.model.predict(test_data_x, verbose=0).reshape(-1)
            predictions.append(prediction)

        if best_model.num_classes > 2:
            difficulity.append(
                multi_difficulty_formula(
                    predictions, test_data_y, NN.num_classes
                ).values
            )

        else:
            difficulity.append(
                binary_difficulty_formula(predictions, test_data_y).values
            )

    return pd.DataFrame(difficulity)


def binary_difficulty_formula(predictions, y_test):
    locations = abs(np.mean(predictions, axis=0) - y_test)
    distribution = np.std(predictions, axis=0) / math.sqrt(1 / 12)
    difficulty = locations + distribution / 2
    return difficulty


def multi_difficulty_formula(predictions, y_test, num_classes):
    one_hot_vector = [0] * num_classes
    one_hot_vector[y_test] = 1
    location = abs(np.mean(predictions, axis=0) - one_hot_vector)
    distribution = np.std(predictions, axis=0) / math.sqrt(1 / 12)
    class_difficulty = (location + distribution) / 2
    difficulty = np.mean(class_difficulty, axis=1)
    return difficulty



import logging
import ray
from ray import tune, train
import itertools
import warnings

warnings.filterwarnings("ignore")


def get_available_resources():
    """_summary_

    :return: _description_
    :rtype: _type_
    """
    if not ray.is_initialized():
        ray.init(logging_level=logging.CRITICAL, log_to_driver=False)
    resource = ray.cluster_resources()
    if ray.is_initialized():
        ray.shutdown()
    return {"CPU": resource.get("CPU"), "GPU": resource.get("GPU")}


def tune_parameters(model, data, target_column, max_layers, max_units, resources):
    """ """
    if not ray.is_initialized():
        ray.init(logging_level=logging.CRITICAL, log_to_driver=False)

    neurons = list(range(1, max_layers + 1))
    layer_neuron_orders = [
        combination
        for r in range(1, max_units)
        for combination in itertools.product(neurons, repeat=r)
    ]
    X = data.drop(target_column, axis=1)
    Y = data[target_column]
    search_space = {
        "learnRate": tune.grid_search([0.01, 0.03, 0.1]),
        "batch_size": tune.grid_search([32, 64, 128]),
        "activation": tune.grid_search(["relu", "tanh"]),
        "number_of_neurons": tune.grid_search(layer_neuron_orders),
        "num_classes": len(set(Y.values)),
        "input_size": len(X.columns),
    }

    trainer_with_resources = tune.with_resources(
        tune.with_parameters(model, data_x=X, data_y=Y), resources=resources
    )
    tuner = tune.Tuner(
        trainer_with_resources,
        param_space=search_space,
        tune_config=tune.TuneConfig(
            metric="mean_accuracy",
            mode="max",
        ),
        run_config=train.RunConfig(verbose=0, log_to_file=False),
    )
    tuning_result = tuner.fit()
    best_config = tuning_result.get_best_result().config

    return best_config
